import os
import sys
import mock
import tempfile
import shutil
import errno

import os.path as op

import mock

from six.moves \
    import \
        StringIO
from bento.compat.api \
    import \
        NamedTemporaryFile
from bento.compat.api.moves \
    import \
        unittest

from bento.utils.utils \
    import subst_vars, to_camel_case, explode_path, same_content, \
        cmd_is_runnable, memoized, comma_list_split, cpu_count, pprint, \
        virtualenv_prefix
from bento.utils.io2 \
    import \
        safe_write
from bento.utils.os2 \
    import \
        rename
import bento.utils.path

def raise_oserror(err):
    ex = OSError()
    ex.errno = err
    raise ex

class TestSubstVars(unittest.TestCase):
    def test_subst_vars_simple(self):
        d = {'prefix': '/usr/local'}
        self.assertEqual(subst_vars('$prefix', d), d['prefix'])

    def test_subst_vars_recursive(self):
        d = {'prefix': '/usr/local',
             'eprefix': '$prefix',
             'datarootdir': '$prefix/share',
             'datadir': '$datarootdir'}
        self.assertEqual(subst_vars('$eprefix', d), '/usr/local')
        self.assertEqual(subst_vars('$datadir', d), '/usr/local/share')

    def test_subst_vars_escaped(self):
        d = {'prefix': '/usr/local',
             'eprefix': '$prefix',
             'datarootdir': '$prefix/share',
             'datadir': '$datarootdir'}
        self.assertEqual(subst_vars('$datadir', d), '/usr/local/share')
        self.assertEqual(subst_vars('$$datadir', d), '$datadir')

    def test_to_camel_case(self):
        d = [("foo", "Foo"), ("foo_bar", "FooBar"), ("_foo_bar", "_FooBar"), ("__fubar", "__Fubar"),
             ("_fubar_", "_Fubar_")]
        for lower, camel in d:
            self.assertEqual(to_camel_case(lower), camel)

class TestExplodePath(unittest.TestCase):
    def test_simple(self):
        path = "/home/joe"
        self.assertEqual(explode_path(path), ["/", "home", "joe"])

        path = "home/joe"
        self.assertEqual(explode_path(path), ["home", "joe"])

    def test_ends_with_sep(self):
        path = "/home/joe/"
        self.assertEqual(explode_path(path), ["/", "home", "joe"])

    def test_empty(self):
        path = ""
        self.assertEqual(explode_path(path), [])

class TestSameFile(unittest.TestCase):
    def test_same(self):
        f1 = NamedTemporaryFile("wt", delete=False)
        try:
            f1.write("fofo")
            f1.close()
            f2 = NamedTemporaryFile("wt", delete=False)
            try:
                f2.write("fofo")
                f2.close()
                self.assertTrue(same_content(f1.name, f2.name))
            finally:
                os.remove(f2.name)
        finally:
            os.remove(f1.name)

    def test_different(self):
        f1 = NamedTemporaryFile("wt", delete=False)
        try:
            f1.write("fofo")
            f1.close()
            f2 = NamedTemporaryFile("wt", delete=False)
            try:
                f2.write("fofa")
                f2.close()
                self.assertFalse(same_content(f1.name, f2.name))
            finally:
                os.remove(f2.name)
        finally:
            os.remove(f1.name)

class TestMisc(unittest.TestCase):
    def test_cmd_is_runnable(self):
        st = cmd_is_runnable(["python", "-c", "''"])
        self.assertTrue(st)

    def test_cmd_is_not_runnable(self):
        st = cmd_is_runnable(["phython", "-c", "''"])
        self.assertFalse(st)

    def test_simple_ensure_dir(self):
        d = tempfile.mkdtemp()
        try:
            new_dir = op.join(d, "foo", "bar")
            new_file = op.join(new_dir, "fubar.txt")
            bento.utils.path.ensure_dir(new_file)
            self.assertTrue(op.exists(new_dir))
        finally:
            shutil.rmtree(d)

    def _test_rename(self):
        d = tempfile.mkdtemp()
        try:
            f = op.join(d, "f.txt")
            fid = open(f, "wt")
            try:
                fid.write("")
            finally:
                fid.close()
            g = op.join(d, "g.txt")
            rename(f, g)
            self.assertTrue(op.exists(g))
        finally:
            shutil.rmtree(d)

    def test_rename(self):
        self._test_rename()

    @mock.patch("bento.utils.os2._rename", lambda x, y: raise_oserror(errno.EXDEV))
    def test_rename_exdev_failure(self):
        self._test_rename()

    @mock.patch("bento.utils.os2._rename", lambda x, y: raise_oserror(errno.EBUSY))
    def test_rename_failure(self):
        self.assertRaises(OSError, self._test_rename)

class TestMemoize(unittest.TestCase):
    def test_simple_no_arguments(self):
        lst = []
        @memoized
        def dummy_function():
            lst.append(1)

        dummy_function()
        dummy_function()

        self.assertEqual(lst, [1])

    def test_simple(self):
        lst = []
        @memoized
        def dummy_function(x):
            lst.append(x)

        dummy_function(1)
        dummy_function(2)
        dummy_function(1)

        self.assertEqual(lst, [1, 2])

    def test_mutable_argument(self):
        lst = []
        @memoized
        def dummy_function(x):
            lst.extend(x)

        dummy_function([1])
        dummy_function([2])
        dummy_function([1])

        self.assertEqual(lst, [1, 2, 1])

    def test_meta(self):
        @memoized
        def dummy_function():
            """some dummy doc"""
        self.assertEqual(repr(dummy_function), "some dummy doc")

    def test_instance_method(self):
        lst = []
        class Foo(object):
            @memoized
            def dummy_function(self, x):
                lst.append(x)
        foo = Foo()
        foo.dummy_function(1)
        foo.dummy_function(1)

        self.assertEqual(lst, [1])

class TestCommaListSplit(unittest.TestCase):
    def test_simple(self):
        self.assertEqual(comma_list_split("-a,-b"), ["-a", "-b"])

class TestPathNormalization(unittest.TestCase):
    @mock.patch("os.sep", "\\")
    def test_simple_win32(self):
        self.assertEqual(bento.utils.path.normalize_path(r"foo\bar"), "foo/bar")
        self.assertEqual(bento.utils.path.unnormalize_path("foo/bar"), r"foo\bar")

    @mock.patch("os.sep", "/")
    def test_simple_unix(self):
        self.assertEqual(bento.utils.path.normalize_path(r"foo\bar"), "foo\\bar")
        self.assertEqual(bento.utils.path.unnormalize_path("foo/bar"), "foo/bar")

class TestPPrint(unittest.TestCase):
    @mock.patch("bento.utils.utils.COLORS_LST", {"USE": True, "RED": "\x1b[01;31m", "NORMAL": "\x1b[0m"})
    def test_simple(self):
        s = StringIO()
        pprint("RED", "foo", s)
        self.assertEqual(s.getvalue(), "\x1b[01;31mfoo\x1b[0m\n")

    @mock.patch("bento.utils.utils.COLORS_LST", {"USE": False})
    def test_simple_no_color(self):
        s = StringIO()
        pprint("RED", "foo", s)
        self.assertEqual(s.getvalue(), "foo\n")

class TestVirtualEnvPrefix(unittest.TestCase):
    @mock.patch("sys.real_prefix", sys.prefix, create=True)
    def test_no_virtualenv(self):
        self.assertEqual(virtualenv_prefix(), None)

    @mock.patch("sys.real_prefix", sys.prefix, create=True)
    @mock.patch("sys.prefix", "yoyo")
    def test_virtualenv(self):
        self.assertEqual(virtualenv_prefix(), "yoyo")

class TestCpuCount(unittest.TestCase):
    def test_native(self):
        self.assertTrue(cpu_count() > 0)

    @mock.patch("sys.platform", "win32")
    def test_win32(self):
        old = os.environ.get("NUMBER_OF_PROCESSORS", None)
        try:
            os.environ["NUMBER_OF_PROCESSORS"] = "2"
            self.assertEqual(cpu_count(), 2)
        finally:
            if old is not None:
                os.environ["NUMBER_OF_PROCESSORS"] = old

    @mock.patch("sys.platform", "bsd")
    @mock.patch("os.popen", lambda s: StringIO("3"))
    def test_bsd(self):
        self.assertEqual(cpu_count(), 3)



import yaku.utils

def setup(ctx):
    env = ctx.env

    ctx.env["CXX"] = ["g++"]
    ctx.env["CXX_TGT_F"] = ["-c", "-o"]
    ctx.env["CXX_SRC_F"] = []
    ctx.env["CXXFLAGS"] = ["-Wall"]
    ctx.env["CXXLINK"] = ["g++"]
    ctx.env["CXXLINKFLAGS"] = []
    ctx.env["CXXLINK_TGT_F"] = ["-o"]
    ctx.env["CXXLINK_SRC_F"] = []
    ctx.env["CXXSHLINK"] = ["g++", "-shared"]
    ctx.env["CXXSHLINKFLAGS"] = []
    ctx.env["CXXSHLINK_TGT_F"] = ["-o"]
    ctx.env["CXXSHLINK_SRC_F"] = []
    ctx.env["CPPPATH"] = []
    ctx.env["CPPPATH_FMT"] = "-I%s"
    ctx.env["LIBDIR"] = []
    ctx.env["LIBS"] = []
    ctx.env["FRAMEWORKS"] = []
    ctx.env["LIB_FMT"] = "-l%s"
    ctx.env["LIBDIR_FMT"] = "-L%s"

    ctx.env["CXX_OBJECT_FMT"] = "%s.o"
    ctx.env["PROGRAM_FMT"] = "%s"

def detect(ctx):
    if yaku.utils.find_program("gcc") is None:
        return False
    else:
        return True



import sys
import re

if sys.version_info[0] < 3:
    from cStringIO \
        import \
            StringIO
else:
    from io \
        import \
            StringIO

from yaku.utils \
    import \
        ensure_dir
from yaku.conftests.conftests \
    import \
       check_compiler, check_type, check_header, check_func, \
       check_lib, check_type_size, define, check_funcs_at_once, check_cpp_symbol

VALUE_SUB = re.compile('[^A-Z0-9_]')

def generate_config_h(conf_res, name):
    def value_to_string(value):
        s = value.upper()
        return VALUE_SUB.sub("_", s)

    def var_name(entry):
        if entry["type"] == "header":
            return "HAVE_%s 1" % entry["value"].upper().replace(".", "_")
        elif entry["type"] == "type":
            return "HAVE_%s 1" % entry["value"].upper().replace(" ", "_")
        elif entry["type"] == "type_size":
            return "SIZEOF_%s %s" % (
                    value_to_string(entry["value"]),
                    entry["result"])
        elif entry["type"] == "lib":
            return "HAVE_LIB%s 1" % entry["value"].upper()
        elif entry["type"] == "func":
            return "HAVE_%s 1" % entry["value"].upper()
        elif entry["type"] == "decl":
            return "HAVE_DECL_%s 1" % entry["value"].upper()
        else:
            raise ValueError("Bug: entry %s not handled" % entry)

    def comment(entry):
        if entry["type"] == "header":
            return r"""
/* Define to 1 if you have the <%s> header file. */
""" % entry["value"]
        elif entry["type"] == "lib":
            return r"""
/* Define to 1 if you have the `%s' library. */
""" % entry["value"]
        elif entry["type"] == "func":
            return r"""
/* Define to 1 if you have the `%s' function. */
""" % entry["value"]
        elif entry["type"] == "decl":
            return r"""
/* Set to 1 if %s is defined. */
""" % entry["value"]
        elif entry["type"] == "type":
            return r"""
/* Define if your compiler provides %s */
""" % entry["value"]
        elif entry["type"] == "type_size":
            return r"""
/* The size of `%s', as computed by sizeof. */
""" % entry["value"]
        else:
            raise ValueError("Bug: entry %s not handled" % entry)

    buf = StringIO()
    for entry in conf_res:
        if entry["type"] == "define":
            buf.write(entry["value"])
        else:
            var = var_name(entry)
            if entry["result"]:
                buf.write(comment(entry))
                buf.write("#define %s\n" % var)

    ensure_dir(name)
    fid = open(name, "w")
    try:
        fid.write(buf.getvalue())
    finally:
        fid.close()



from bento.compat.api.moves \
    import \
        unittest

from bento.commands.dependency \
    import \
        CommandScheduler

class TestCommandScheduler(unittest.TestCase):
    def test_simple(self):
        scheduler = CommandScheduler()
        scheduler.set_before("task2", "task1")
        scheduler.set_after("task2", "task3")
        scheduler.set_after("task2", "task4")
        scheduler.set_before("task4", "task3")

        tasks = scheduler.order("task4")
        self.assertEqual(tasks, ["task1", "task2", "task3"])

    def test_cycle(self):
        scheduler = CommandScheduler()
        scheduler.set_before("task2", "task1")
        scheduler.set_before("task1", "task2")

        self.assertRaises(ValueError, lambda: scheduler.order("task1"))



"""Subset of inspect module from upstream python

We use this instead of upstream because upstream inspect is slow to import, and
significanly contributes to numpy import times. Importing this copy has almost
no overhead.
"""

import os
import sys
import types
import string
import imp
import linecache
import re

__all__ = ['getargspec', 'formatargspec']

# ----------------------------------------------------------- type-checking
def ismodule(object):
    """Return true if the object is a module.

    Module objects provide these attributes:
        __doc__         documentation string
        __file__        filename (missing for built-in modules)"""
    return isinstance(object, types.ModuleType)

def isclass(object):
    """Return true if the object is a class.

    Class objects provide these attributes:
        __doc__         documentation string
        __module__      name of module in which this class was defined"""
    if sys.version_info[0] < 3:
        return isinstance(object, types.ClassType) or hasattr(object, '__bases__')
    else:
        return isinstance(object, type) or hasattr(object, '__bases__')

def ismethod(object):
    """Return true if the object is an instance method.

    Instance method objects provide these attributes:
        __doc__         documentation string
        __name__        name with which this method was defined
        im_class        class object in which this method belongs
        im_func         function object containing implementation of method
        im_self         instance to which this method is bound, or None"""
    return isinstance(object, types.MethodType)

def isfunction(object):
    """Return true if the object is a user-defined function.

    Function objects provide these attributes:
        __doc__         documentation string
        __name__        name with which this function was defined
        func_code       code object containing compiled function bytecode
        func_defaults   tuple of any default values for arguments
        func_doc        (same as __doc__)
        func_globals    global namespace in which this function was defined
        func_name       (same as __name__)"""
    return isinstance(object, types.FunctionType)

def istraceback(object):
    """Return true if the object is a traceback.

    Traceback objects provide these attributes:
        tb_frame        frame object at this level
        tb_lasti        index of last attempted instruction in bytecode
        tb_lineno       current line number in Python source code
        tb_next         next inner traceback object (called by this level)"""
    return isinstance(object, types.TracebackType)

def isframe(object):
    """Return true if the object is a frame object.

    Frame objects provide these attributes:
        f_back          next outer frame object (this frame's caller)
        f_builtins      built-in namespace seen by this frame
        f_code          code object being executed in this frame
        f_exc_traceback traceback if raised in this frame, or None
        f_exc_type      exception type if raised in this frame, or None
        f_exc_value     exception value if raised in this frame, or None
        f_globals       global namespace seen by this frame
        f_lasti         index of last attempted instruction in bytecode
        f_lineno        current line number in Python source code
        f_locals        local namespace seen by this frame
        f_restricted    0 or 1 if frame is in restricted execution mode
        f_trace         tracing function for this frame, or None"""
    return isinstance(object, types.FrameType)

def iscode(object):
    """Return true if the object is a code object.

    Code objects provide these attributes:
        co_argcount     number of arguments (not including * or ** args)
        co_code         string of raw compiled bytecode
        co_consts       tuple of constants used in the bytecode
        co_filename     name of file in which this code object was created
        co_firstlineno  number of first line in Python source code
        co_flags        bitmap: 1=optimized | 2=newlocals | 4=*arg | 8=**arg
        co_lnotab       encoded mapping of line numbers to bytecode indices
        co_name         name with which this code object was defined
        co_names        tuple of names of local variables
        co_nlocals      number of local variables
        co_stacksize    virtual machine stack space required
        co_varnames     tuple of names of arguments and local variables"""
    return isinstance(object, types.CodeType)

# ------------------------------------------------ argument list extraction
# These constants are from Python's compile.h.
CO_OPTIMIZED, CO_NEWLOCALS, CO_VARARGS, CO_VARKEYWORDS = 1, 2, 4, 8

def getargs(co):
    """Get information about the arguments accepted by a code object.

    Three things are returned: (args, varargs, varkw), where 'args' is
    a list of argument names (possibly containing nested lists), and
    'varargs' and 'varkw' are the names of the * and ** arguments or None."""

    if not iscode(co):
        raise TypeError('arg is not a code object')

    code = co.co_code
    nargs = co.co_argcount
    names = co.co_varnames
    args = list(names[:nargs])
    step = 0

    # The following acrobatics are for anonymous (tuple) arguments.
    for i in range(nargs):
        if args[i][:1] in ['', '.']:
            stack, remain, count = [], [], []
            while step < len(code):
                op = ord(code[step])
                step = step + 1
                if op >= dis.HAVE_ARGUMENT:
                    opname = dis.opname[op]
                    value = ord(code[step]) + ord(code[step+1])*256
                    step = step + 2
                    if opname in ['UNPACK_TUPLE', 'UNPACK_SEQUENCE']:
                        remain.append(value)
                        count.append(value)
                    elif opname == 'STORE_FAST':
                        stack.append(names[value])

                        # Special case for sublists of length 1: def foo((bar))
                        # doesn't generate the UNPACK_TUPLE bytecode, so if
                        # `remain` is empty here, we have such a sublist.
                        if not remain:
                            stack[0] = [stack[0]]
                            break
                        else:
                            remain[-1] = remain[-1] - 1
                            while remain[-1] == 0:
                                remain.pop()
                                size = count.pop()
                                stack[-size:] = [stack[-size:]]
                                if not remain: break
                                remain[-1] = remain[-1] - 1
                            if not remain: break
            args[i] = stack[0]

    varargs = None
    if co.co_flags & CO_VARARGS:
        varargs = co.co_varnames[nargs]
        nargs = nargs + 1
    varkw = None
    if co.co_flags & CO_VARKEYWORDS:
        varkw = co.co_varnames[nargs]
    return args, varargs, varkw

def getargspec(func):
    """Get the names and default values of a function's arguments.

    A tuple of four things is returned: (args, varargs, varkw, defaults).
    'args' is a list of the argument names (it may contain nested lists).
    'varargs' and 'varkw' are the names of the * and ** arguments or None.
    'defaults' is an n-tuple of the default values of the last n arguments.
    """

    if ismethod(func):
        func = func.im_func
    if not isfunction(func):
        raise TypeError('arg is not a Python function')
    args, varargs, varkw = getargs(func.func_code)
    return args, varargs, varkw, func.func_defaults

def getargvalues(frame):
    """Get information about arguments passed into a particular frame.

    A tuple of four things is returned: (args, varargs, varkw, locals).
    'args' is a list of the argument names (it may contain nested lists).
    'varargs' and 'varkw' are the names of the * and ** arguments or None.
    'locals' is the locals dictionary of the given frame."""
    args, varargs, varkw = getargs(frame.f_code)
    return args, varargs, varkw, frame.f_locals

def joinseq(seq):
    if len(seq) == 1:
        return '(' + seq[0] + ',)'
    else:
        return '(' + ', '.join(seq) + ')'

def strseq(object, convert, join=joinseq):
    """Recursively walk a sequence, stringifying each element."""
    if type(object) in [types.ListType, types.TupleType]:
        return join(map(lambda o, c=convert, j=join: strseq(o, c, j), object))
    else:
        return convert(object)

def formatargspec(args, varargs=None, varkw=None, defaults=None,
                  formatarg=str,
                  formatvarargs=lambda name: '*' + name,
                  formatvarkw=lambda name: '**' + name,
                  formatvalue=lambda value: '=' + repr(value),
                  join=joinseq):
    """Format an argument spec from the 4 values returned by getargspec.

    The first four arguments are (args, varargs, varkw, defaults).  The
    other four arguments are the corresponding optional formatting functions
    that are called to turn names and values into strings.  The ninth
    argument is an optional function to format the sequence of arguments."""
    specs = []
    if defaults:
        firstdefault = len(args) - len(defaults)
    for i in range(len(args)):
        spec = strseq(args[i], formatarg, join)
        if defaults and i >= firstdefault:
            spec = spec + formatvalue(defaults[i - firstdefault])
        specs.append(spec)
    if varargs is not None:
        specs.append(formatvarargs(varargs))
    if varkw is not None:
        specs.append(formatvarkw(varkw))
    return '(' + ', '.join(specs) + ')'

def formatargvalues(args, varargs, varkw, locals,
                    formatarg=str,
                    formatvarargs=lambda name: '*' + name,
                    formatvarkw=lambda name: '**' + name,
                    formatvalue=lambda value: '=' + repr(value),
                    join=joinseq):
    """Format an argument spec from the 4 values returned by getargvalues.

    The first four arguments are (args, varargs, varkw, locals).  The
    next four arguments are the corresponding optional formatting functions
    that are called to turn names and values into strings.  The ninth
    argument is an optional function to format the sequence of arguments."""
    def convert(name, locals=locals,
                formatarg=formatarg, formatvalue=formatvalue):
        return formatarg(name) + formatvalue(locals[name])
    specs = []
    for i in range(len(args)):
        specs.append(strseq(args[i], convert, join))
    if varargs:
        specs.append(formatvarargs(varargs) + formatvalue(locals[varargs]))
    if varkw:
        specs.append(formatvarkw(varkw) + formatvalue(locals[varkw]))
    return '(' + string.join(specs, ', ') + ')'

# -------------------------------------------------- source code extraction
def indentsize(line):
    """Return the indent size, in spaces, at the start of a line of text."""
    expline = string.expandtabs(line)
    return len(expline) - len(string.lstrip(expline))

def getdoc(object):
    """Get the documentation string for an object.

    All tabs are expanded to spaces.  To clean up docstrings that are
    indented to line up with blocks of code, any whitespace than can be
    uniformly removed from the second line onwards is removed."""
    try:
        doc = object.__doc__
    except AttributeError:
        return None
    if not isinstance(doc, types.StringTypes):
        return None
    try:
        lines = string.split(string.expandtabs(doc), '\n')
    except UnicodeError:
        return None
    else:
        # Find minimum indentation of any non-blank lines after first line.
        margin = sys.maxint
        for line in lines[1:]:
            content = len(string.lstrip(line))
            if content:
                indent = len(line) - content
                margin = min(margin, indent)
        # Remove indentation.
        if lines:
            lines[0] = lines[0].lstrip()
        if margin < sys.maxint:
            for i in range(1, len(lines)): lines[i] = lines[i][margin:]
        # Remove any trailing or leading blank lines.
        while lines and not lines[-1]:
            lines.pop()
        while lines and not lines[0]:
            lines.pop(0)
        return string.join(lines, '\n')

def getfile(object):
    """Work out which source or compiled file an object was defined in."""
    if ismodule(object):
        if hasattr(object, '__file__'):
            return object.__file__
        raise TypeError('arg is a built-in module')
    if isclass(object):
        object = sys.modules.get(object.__module__)
        if hasattr(object, '__file__'):
            return object.__file__
        raise TypeError('arg is a built-in class')
    if ismethod(object):
        object = object.im_func
    if isfunction(object):
        object = object.func_code
    if istraceback(object):
        object = object.tb_frame
    if isframe(object):
        object = object.f_code
    if iscode(object):
        return object.co_filename
    raise TypeError('arg is not a module, class, method, '
                    'function, traceback, frame, or code object')

def findsource(object):
    """Return the entire source file and starting line number for an object.

    The argument may be a module, class, method, function, traceback, frame,
    or code object.  The source code is returned as a list of all the lines
    in the file and the line number indexes a line in that list.  An IOError
    is raised if the source code cannot be retrieved."""
    file = getsourcefile(object) or getfile(object)
    lines = linecache.getlines(file)
    if not lines:
        raise IOError('could not get source code')

    if ismodule(object):
        return lines, 0

    if isclass(object):
        name = object.__name__
        pat = re.compile(r'^\s*class\s*' + name + r'\b')
        for i in range(len(lines)):
            if pat.match(lines[i]): return lines, i
        else:
            raise IOError('could not find class definition')

    if ismethod(object):
        object = object.im_func
    if isfunction(object):
        object = object.func_code
    if istraceback(object):
        object = object.tb_frame
    if isframe(object):
        object = object.f_code
    if iscode(object):
        if not hasattr(object, 'co_firstlineno'):
            raise IOError('could not find function definition')
        lnum = object.co_firstlineno - 1
        pat = re.compile(r'^(\s*def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
        while lnum > 0:
            if pat.match(lines[lnum]): break
            lnum = lnum - 1
        return lines, lnum
    raise IOError('could not find code object')

def getmoduleinfo(path):
    """Get the module name, suffix, mode, and module type for a given file."""
    filename = os.path.basename(path)
    def _f(suffix, mode, mtype):
       return (-len(suffix), suffix, mode, mtype)
    suffixes = map(_f, imp.get_suffixes())
    suffixes.sort() # try longest suffixes first, in case they overlap
    for neglen, suffix, mode, mtype in suffixes:
        if filename[neglen:] == suffix:
            return filename[:neglen], suffix, mode, mtype

def getmodulename(path):
    """Return the module name for a given file, or None."""
    info = getmoduleinfo(path)
    if info: return info[0]

def getsourcefile(object):
    """Return the Python source file an object was defined in, if it exists."""
    filename = getfile(object)
    if filename[-4:].lower() in ['.pyc', '.pyo']:
        filename = filename[:-4] + '.py'
    for suffix, mode, kind in imp.get_suffixes():
        if 'b' in mode and filename[-len(suffix):].lower() == suffix:
            # Looks like a binary file.  We want to only return a text file.
            return None
    if os.path.exists(filename):
        return filename

# -------------------------------------------------- stack frame extraction
def getframeinfo(frame, context=1):
    """Get information about a frame or traceback object.

    A tuple of five things is returned: the filename, the line number of
    the current line, the function name, a list of lines of context from
    the source code, and the index of the current line within that list.
    The optional second argument specifies the number of lines of context
    to return, which are centered around the current line."""
    if istraceback(frame):
        lineno = frame.tb_lineno
        frame = frame.tb_frame
    else:
        lineno = frame.f_lineno
    if not isframe(frame):
        raise TypeError('arg is not a frame or traceback object')

    filename = getsourcefile(frame) or getfile(frame)
    if context > 0:
        start = lineno - 1 - context//2
        try:
            lines, lnum = findsource(frame)
        except IOError:
            lines = index = None
        else:
            start = max(start, 1)
            start = max(0, min(start, len(lines) - context))
            lines = lines[start:start+context]
            index = lineno - 1 - start
    else:
        lines = index = None

    return (filename, lineno, frame.f_code.co_name, lines, index)

def getlineno(frame):
    """Get the line number from a frame object, allowing for optimization."""
    # FrameType.f_lineno is now a descriptor that grovels co_lnotab
    return frame.f_lineno

def getouterframes(frame, context=1):
    """Get a list of records for a frame and all higher (calling) frames.

    Each record contains a frame object, filename, line number, function
    name, a list of lines of context, and index within the context."""
    framelist = []
    while frame:
        framelist.append((frame,) + getframeinfo(frame, context))
        frame = frame.f_back
    return framelist

def getinnerframes(tb, context=1):
    """Get a list of records for a traceback's frame and all lower frames.

    Each record contains a frame object, filename, line number, function
    name, a list of lines of context, and index within the context."""
    framelist = []
    while tb:
        framelist.append((tb.tb_frame,) + getframeinfo(tb, context))
        tb = tb.tb_next
    return framelist

currentframe = sys._getframe

def stack(context=1):
    """Return a list of records for the stack above the caller's frame."""
    return getouterframes(sys._getframe(1), context)

if __name__ == '__main__':
    import inspect
    def foo(x, y, z=None):
        return None

    print(inspect.getargs(foo.func_code))
    print(getargs(foo.func_code))

    print(inspect.getargspec(foo))
    print(getargspec(foo))

    print(inspect.formatargspec(*inspect.getargspec(foo)))
    print(formatargspec(*getargspec(foo)))



import os
import sys
import pickle
import tempfile
import shutil
import copy

import os.path as op

from bento.compat.api.moves \
    import \
        unittest
from bento.core.node \
    import \
        Node, create_root_with_source_tree, find_root, split_path_win32, split_path_cygwin

class TestNode(unittest.TestCase):
    def setUp(self):
        self.root = Node("", None)

        if sys.version_info[0] < 3:
            self._string_classes = [str, unicode]
        else:
            self._string_classes = [str]

    def test_scratch_creation(self):
        root = Node("", None)
        self.assertEqual(root, root)

    def test_find_node(self):
        r_n = os.path.abspath(os.getcwd())

        for f in self._string_classes:
            n = self.root.find_node(f(os.getcwd()))
            self.assertTrue(n)
            assert n.abspath() == r_n

    def test_make_node(self):
        r_n = os.path.abspath(os.getcwd())

        for f in self._string_classes:
            n = self.root.make_node(f(os.getcwd()))
            self.assertTrue(n)
            assert n.abspath() == r_n, "%s vs %s" % (n.abspath(), r_n)

    def test_serialization(self):
        """Test pickled/unpickle round trip."""
        n = self.root.find_node(os.getcwd())
        r_n = pickle.loads(pickle.dumps(n))
        self.assertEqual(n.name, r_n.name)
        self.assertEqual(n.abspath(), r_n.abspath())
        self.assertEqual(n.parent.abspath(), r_n.parent.abspath())
        self.assertEqual([child.abspath() for child in getattr(n, "children", [])],
                         [child.abspath() for child in getattr(n, "children", [])])

    def test_str_repr(self):
        d = tempfile.mkdtemp()
        try:
            r_n = os.path.abspath(os.path.join(d, "foo.txt"))
            node = self.root.make_node(r_n)
            self.assertEqual(str(node), "foo.txt")
            self.assertEqual(repr(node), r_n)
        finally:
            shutil.rmtree(d)

    def test_invalid_copy(self):
        self.assertRaises(NotImplementedError, lambda: copy.copy(self.root))

class TestNodeInsideTemp(unittest.TestCase):
    def setUp(self):
        root = Node("", None)
        self.d_node = root.find_node(tempfile.mkdtemp())
        assert self.d_node is not None

    def tearDown(self):
        shutil.rmtree(self.d_node.abspath())

    def test_delete(self):
        n = self.d_node.make_node("foo.txt")
        n.write("foo")
        self.assertEqual(n.read(), "foo")
        n.delete()
        self.assertEqual(self.d_node.listdir(), [])

    def test_delete_dir(self):
        foo_node = self.d_node.make_node("foo")
        n = foo_node.make_node("bar.txt")
        n.parent.mkdir()
        n.write("foo")
        self.assertEqual(n.read(), "foo")
        foo_node.delete()
        self.assertEqual(self.d_node.listdir(), [])

    def test_mkdir(self):
        bar_node = self.d_node.make_node(op.join("foo", "bar"))
        os.makedirs(bar_node.parent.abspath())
        os.makedirs(bar_node.abspath())
        bar_node.mkdir()

    def test_suffix(self):
        foo = self.d_node.make_node("foo.txt")
        self.assertEqual(foo.suffix(), ".txt")

        foo = self.d_node.make_node("foo.txt.swp")
        self.assertEqual(foo.suffix(), ".swp")

    def test_make_node(self):
        foo = self.d_node.make_node("foo")
        foo.make_node("../bar")
        self.assertTrue(self.d_node.find_node("bar") is not None)

    def test_ant(self):
        for filename in ["bar.txt", "foo.bar", "fubar.txt"]:
            n = self.d_node.make_node(filename)
            n.write("")
        nodes = self.d_node.ant_glob("*.txt")
        bar = self.d_node.find_node("bar.txt")
        fubar = self.d_node.find_node("fubar.txt")
        self.assertEqual(set([fubar.abspath(), bar.abspath()]), set([node.abspath() for node in nodes]))

    def test_ant_excl(self):
        for filename in ["bar.txt", "foo.bar", "fubar.txt"]:
            n = self.d_node.make_node(filename)
            n.write("")
        nodes = self.d_node.ant_glob("*", excl=["*.txt"])
        foobar = self.d_node.find_node("foo.bar")
        self.assertEqual(set(node.abspath() for node in nodes), set([foobar.abspath()]))

class TestNodeWithBuild(unittest.TestCase):
    def setUp(self):
        top = os.getcwd()
        build = os.path.join(os.getcwd(), "_tmp_build")

        self.root = create_root_with_source_tree(top, build)

    def test_root(self):
        self.assertEqual(self.root, find_root(self.root))

    def test_cwd_node(self):
        cur_node = self.root.make_node(os.getcwd())
        self.assertEqual(os.getcwd(), cur_node.abspath())

class TestUtils(unittest.TestCase):
    def test_split_path_win32(self):
        self.assertEqual(split_path_win32(r"C:\foo\bar"), ["C:", "foo", "bar"])
        self.assertEqual(split_path_win32(r"\\"), ["\\"])
        self.assertEqual(split_path_win32(""), [''])

    def test_split_path_cygwin(self):
        self.assertEqual(split_path_cygwin("/foo/bar"), ["", "foo", "bar"])
        self.assertEqual(split_path_cygwin("//"), ["/"])
        self.assertEqual(split_path_cygwin(""), [''])



import os
import copy

from yaku.task_manager \
    import \
        extension, CompiledTaskGen
from yaku.task \
    import \
        task_factory
from yaku.compiled_fun \
    import \
        compile_fun
from yaku.utils \
    import \
        ensure_dir, get_exception
from yaku.tools.ctasks \
    import \
        apply_libdir
from yaku.errors \
    import \
        TaskRunFailure
from yaku.scheduler \
    import \
        run_tasks
from yaku.conf \
    import \
        with_conf_blddir, create_file, write_log
import yaku.tools

f77_compile, f77_vars = compile_fun("f77", "${F77} ${F77FLAGS} ${F77_TGT_F}${TGT[0].abspath()} ${F77_SRC_F}${SRC}", False)

fprogram, fprogram_vars = compile_fun("fprogram", "${F77_LINK} ${F77_LINK_SRC_F}${SRC} ${F77_LINK_TGT_F}${TGT[0].abspath()} ${APP_LIBDIR} ${F77_LINKFLAGS}", False)

@extension(".f")
def fortran_task(self, node):
    tasks = fcompile_task(self, node)
    return tasks

def fcompile_task(self, node):
    base = self.env["F77_OBJECT_FMT"] % node.name
    target = node.parent.declare(base)
    ensure_dir(target.abspath())

    task = task_factory("f77")(inputs=[node], outputs=[target])
    task.gen = self
    task.env_vars = f77_vars
    task.env = self.env
    task.func = f77_compile
    self.object_tasks.append(task)
    return [task]

class FortranBuilder(yaku.tools.Builder):
    def __init__(self, ctx):
        yaku.tools.Builder.__init__(self, ctx)

    def _try_task_maker(self, task_maker, name, body):
        conf = self.ctx
        code =  body
        sources = [create_file(conf, code, name, ".f")]

        task_gen = CompiledTaskGen("conf", conf, sources, name)
        task_gen.env.update(copy.deepcopy(conf.env))

        tasks = task_maker(task_gen, name)
        self.ctx.last_task = tasks[-1]

        for t in tasks:
            t.disable_output = True
            t.log = conf.log

        succeed = False
        explanation = None
        try:
            run_tasks(conf, tasks)
            succeed = True
        except TaskRunFailure:
            e = get_exception()
            explanation = str(e)

        write_log(conf, conf.log, tasks, code, succeed, explanation)
        return succeed

    def try_compile(self, name, body):
        # FIXME: temporary workaround for cyclic import between ctasks and conf
        #from yaku.conf import with_conf_blddir
        return with_conf_blddir(self.ctx, name, body,
                                lambda : self._try_task_maker(self._compile, name, body))

    def compile(self, name, sources, env=None):
        _env = copy.deepcopy(self.env)
        if env is not None:
            _env.update(env)

        task_gen = CompiledTaskGen("fcompile", self.ctx,
                                   sources, name)
        task_gen.env = _env
        tasks = self._compile(name, sources)
        self.ctx.tasks.extend(tasks)

        outputs = []
        for t in tasks:
            outputs.extend(t.outputs)
        return outputs

    def _compile(self, task_gen, name):
        tasks = task_gen.process()
        for t in tasks:
            t.env = task_gen.env
        return tasks

    def try_static_library(self, name, body):
        # FIXME: temporary workaround for cyclic import between ctasks and conf
        #from yaku.conf import with_conf_blddir
        cc_builder = self.ctx.builders["ctasks"]
        return with_conf_blddir(self.ctx, name, body,
                                lambda : self._try_task_maker(cc_builder._static_library, name, body))

    def try_program(self, name, body):
        # FIXME: temporary workaround for cyclic import between ctasks and conf
        #from yaku.conf import with_conf_blddir
        return with_conf_blddir(self.ctx, name, body,
                                lambda : self._try_task_maker(self._program, name, body))

    def program(self, name, sources, env=None):
        _env = copy.deepcopy(self.env)
        if env is not None:
            _env.update(env)

        sources = [self.ctx.src_root.find_resource(s) for s in sources]
        task_gen = CompiledTaskGen("fprogram", self.ctx,
                                   sources, name)
        task_gen.env = _env
        tasks = self._program(task_gen, name)

        self.ctx.tasks.extend(tasks)

        outputs = []
        for t in task_gen.link_task:
            outputs.extend(t.outputs)
        return outputs

    def _program(self, task_gen, name):
        apply_libdir(task_gen)

        tasks = task_gen.process()
        ltask = fprogram_task(task_gen, name)
        tasks.extend(ltask)
        for t in tasks:
            t.env = task_gen.env
        task_gen.link_task = ltask
        return tasks

    def configure(self, candidates=None):
        ctx = self.ctx
        if candidates is None:
            compiler_type = "default"
        else:
            compiler_type = candidates[0]

        if compiler_type == "default":
            fc_type = None
            for tool in ["gfortran", "g77"]:
                fc = ctx.load_tool(tool)
                if fc.detect(ctx):
                    fc_type = tool
            if fc_type is None:
                raise ValueError("No fortran compiler found")
        else:
            fc_type = compiler_type
        fc = ctx.load_tool(fc_type)
        fc.setup(ctx)
        self.configured = True

def fprogram_task(self, name):
    objects = [tsk.outputs[0] for tsk in self.object_tasks]
    def declare_target():
        folder, base = os.path.split(name)
        tmp = folder + os.path.sep + self.env["F77_PROGRAM_FMT"] % base
        return self.bld.path.declare(tmp)
    target = declare_target()
    ensure_dir(target.abspath())

    task = task_factory("fprogram")(inputs=objects, outputs=[target])
    task.gen = self
    task.env = self.env
    task.func = fprogram
    task.env_vars = fprogram_vars
    return [task]

def get_builder(ctx):
    return FortranBuilder(ctx)

def mangler(name, under, double_under, case):
    return getattr(name, case)() + under + (name.replace("_", double_under) and double_under)



import sys
import errno
import warnings

import os.path as op

import ply.yacc

from bento._config \
    import \
        _PICKLED_PARSETAB, _OPTIMIZE_LEX, _DEBUG_YACC
from bento.utils.utils \
    import \
        extract_exception
from bento.errors \
    import \
        InternalBentoError, BentoError, ParseError
from bento.parser.lexer \
    import \
        MyLexer, tokens as _tokens
from bento.parser.nodes \
    import \
        Node

# Do not remove: this is used by PLY
tokens = _tokens

def _has_parser_changed(picklefile):
    # FIXME: private function to determine whether ply will need to write to
    # the pickled grammar file. Highly implementation dependent.
    from ply.yacc import PlyLogger, get_caller_module_dict, ParserReflect, YaccError, LRTable
    errorlog = PlyLogger(sys.stderr)

    pdict = get_caller_module_dict(2)

    # Collect parser information from the dictionary
    pinfo = ParserReflect(pdict, log=errorlog)
    pinfo.get_all()

    if pinfo.error:
        raise YaccError("Unable to build parser")

    # Check signature against table files (if any)
    signature = pinfo.signature()

    # Read the tables
    try:
        lr = LRTable()
        read_signature = lr.read_pickle(picklefile)
        return read_signature != signature
    except Exception:
        return True

class Parser(object):
    def __init__(self, lexer=None):
        if lexer is None:
            self.lexer = MyLexer(stage="post_processed", optimize=_OPTIMIZE_LEX)
        else:
            self.lexer = lexer

        picklefile = _PICKLED_PARSETAB
        if not op.exists(picklefile):
            try:
                fid = open(picklefile, "wb")
                fid.close()
            except IOError:
                e = extract_exception()
                raise BentoError("Could not write pickle file %r (original error was %r)" % (picklefile, e))
        else:
            try:
                fid = open(_PICKLED_PARSETAB, "wb")
                fid.close()
            except IOError:
                # In case of read-only support (no write access, zip import,
                # etc...)
                e = extract_exception()
                if e.errno == errno.EACCES:
                    if _has_parser_changed(_PICKLED_PARSETAB):
                        raise BentoError("Cannot write new updated grammar to file %r" % _PICKLED_PARSETAB)
                else:
                    raise
        self.parser = ply.yacc.yacc(start="stmt_list",
                                picklefile=picklefile,
                                debug=_DEBUG_YACC)

    def parse(self, data):
        res = self.parser.parse(data, lexer=self.lexer)
        # FIXME: this is stupid, deal correctly with empty ast in the grammar proper
        if res is None:
            res = Node("empty")
        return res

    def reset(self):
        # XXX: implements reset for lexer
        self.lexer = MyLexer(stage="post_processed", optimize=_OPTIMIZE_LEX)
        # XXX: ply parser.reset method expects those attributes to
        # exist
        self.parser.statestack = []
        self.parser.symstack = []
        return self.parser.restart()

__PARSER = None
def parse(data):
    global __PARSER
    if __PARSER is None:
        __PARSER = Parser()
    else:
        __PARSER.reset()
    return __PARSER.parse(data)
    #return Parser().parse(data)

#-------------
#   Grammar
#-------------
def p_stmt_list(p):
    """stmt_list : stmt_list stmt"""
    p[0] = p[1]
    if p[2].type not in ("newline",):
        p[0].children.append(p[2])

def p_stmt_list_term(p):
    """stmt_list : stmt"""
    p[0] = Node("stmt_list", children=[p[1]])

def p_stmt_list_empty(p):
    """stmt_list : empty"""
    pass

def p_stmt_list_newline(p):
    """stmt_list : NEWLINE"""
    pass

def p_stmt(p):
    """stmt : meta_stmt
            | library
            | exec
            | path
            | flag
            | extra_source_files
            | data_files
            | dummy
    """
    p[0] = p[1]

def p_empty(p):
    """empty : """
    pass

#----------------
#   Meta data
#----------------
def p_meta_stmt(p):
    """meta_stmt : meta_name_stmt
                 | meta_summary_stmt
                 | meta_description_stmt
                 | meta_description_from_file_stmt
                 | meta_version_stmt
                 | meta_url_stmt
                 | meta_download_url_stmt
                 | meta_author_stmt
                 | meta_author_email_stmt
                 | meta_maintainer_stmt
                 | meta_maintainer_email_stmt
                 | meta_license_stmt
                 | meta_platforms_stmt
                 | meta_keywords_stmt
                 | meta_classifiers_stmt
                 | meta_hook_file_stmt
                 | meta_config_py_stmt
                 | meta_meta_template_files_stmt
                 | meta_subento_stmt
                 | meta_use_backends_stmt
    """
    p[0] = p[1]

def p_meta_name_stmt(p):
    """meta_name_stmt : NAME_ID COLON WORD
    """
    p[0] = Node("name", value=p[3])

def p_meta_summary_stmt(p):
    """meta_summary_stmt : SUMMARY_ID COLON single_line_value
    """
    p[0] = Node("summary", value=p[3])

def p_meta_url_stmt(p):
    """meta_url_stmt : URL_ID COLON anyword
    """
    p[0] = Node("url", value=p[3].value)

def p_meta_download_url_stmt(p):
    """meta_download_url_stmt : DOWNLOAD_URL_ID COLON anyword
    """
    p[0] = Node("download_url", value=p[3].value)

def p_meta_author_stmt(p):
    """meta_author_stmt : AUTHOR_ID COLON single_line_value
    """
    p[0] = Node("author", value=p[3])

def p_meta_author_email_stmt(p):
    """meta_author_email_stmt : AUTHOR_EMAIL_ID COLON anyword
    """
    p[0] = Node("author_email", value=p[3].value)

def p_meta_maintainer_stmt(p):
    """meta_maintainer_stmt : MAINTAINER_ID COLON single_line_value
    """
    p[0] = Node("maintainer", value=p[3])

def p_meta_maintainer_email_stmt(p):
    """meta_maintainer_email_stmt : MAINTAINER_EMAIL_ID COLON anyword
    """
    p[0] = Node("maintainer_email", value=p[3].value)

def p_meta_license_stmt(p):
    """meta_license_stmt : LICENSE_ID COLON anyword
    """
    p[0] = Node("license", value=p[3].value)

def p_meta_description_from_file_stmt(p):
    """meta_description_from_file_stmt : DESCRIPTION_FROM_FILE_ID COLON anyword"""
    p[0] = Node("description_from_file", value=p[3].value)

def p_meta_platforms_stmt(p):
    """meta_platforms_stmt : PLATFORMS_ID COLON comma_list
    """
    p[0] = Node("platforms", value=p[3].value)

def p_meta_keywords_stmt(p):
    """meta_keywords_stmt : KEYWORDS_ID COLON comma_list
    """
    p[0] = Node("keywords", value=p[3].value)

def p_meta_version_stmt(p):
    """meta_version_stmt : VERSION_ID COLON version
    """
    p[0] = Node("version", value=p[3].value)

def p_meta_config_py_stmt(p):
    """meta_config_py_stmt : CONFIG_PY_ID COLON word
    """
    p[0] = Node("config_py", value=p[3].value)

def p_meta_meta_template_file_stmt(p):
    """meta_meta_template_files_stmt : META_TEMPLATE_FILE_ID COLON word
    """
    warnings.warn("MetaTemplateFile field is obsolete - use MetaTemplateFiles itself")
    p[0] = Node("meta_template_files", value=[p[3].value])

def p_meta_meta_template_files_stmt(p):
    """meta_meta_template_files_stmt : META_TEMPLATE_FILES_ID COLON comma_list
    """
    p[0] = Node("meta_template_files", value=p[3].value)

def p_meta_classifiers_stmt(p):
    """meta_classifiers_stmt : CLASSIFIERS_ID COLON classifiers_list"""
    p[0] = Node("classifiers", value=p[3].value)

def p_classifiers_list(p):
    """classifiers_list : indented_classifiers_list
    """
    p[0] = Node("classifiers", value=p[1].value)

def p_classifiers_list_term(p):
    """classifiers_list : classifiers
    """
    p[0] = Node("classifiers", value=p[1].value)

def p_indented_comma_list1(p):
    """indented_classifiers_list : classifiers COMMA INDENT classifiers DEDENT
    """
    p[0] = p[1]
    p[0].value.extend(p[4].value)

def p_indented_comma_list2(p):
    """indented_classifiers_list : INDENT classifiers DEDENT
    """
    p[0] = p[2]

def p_classifiers(p):
    """classifiers : classifiers COMMA classifier"""
    p[0] = p[1]
    p[0].value.append(p[3].value)

def p_classifiers_term(p):
    """classifiers : classifier"""
    p[0] = Node("classifiers", value=[p[1].value])

def p_classifier(p):
    """classifier : classifier classifier_atom"""
    p[0] = Node("classifier", value=(p[1].value + p[2].value))

def p_classifier_term(p):
    """classifier : classifier_atom"""
    p[0] = Node("classifier", value=p[1].value)

def p_classifier_atom(p):
    """classifier_atom : WORD
                       | WS
                       | DOUBLE_COLON
    """
    p[0] = Node("classifier", value=p[1])

def p_dummy(p):
    """dummy : AND BACKSLASH"""
    pass

def p_meta_subento_stmt(p):
    """meta_subento_stmt : SUBENTO_ID COLON comma_list"""
    p[0] = Node("subento", value=p[3].value)

def p_meta_use_backends_stmt(p):
    """meta_use_backends_stmt : USE_BACKENDS_ID COLON comma_list"""
    p[0] = Node("use_backends", value=p[3].value)

def p_meta_hook_file_stmt(p):
    """meta_hook_file_stmt : HOOK_FILE_ID COLON comma_list
    """
    p[0] = Node("hook_files", value=p[3].value)

#---------------------------------------
# Data files and extra sources handling
#---------------------------------------
def p_extra_source_files(p):
    """extra_source_files : EXTRA_SOURCE_FILES_ID COLON comma_list"""
    p[0] = Node("extra_source_files", value=p[3].value)
 
def p_data_files(p):
    """data_files : data_files_declaration INDENT data_files_stmts DEDENT
    """
    p[0] = Node("data_files", children=[p[1]])
    p[0].children.append(p[3])

def p_data_files_declaration(p):
    """data_files_declaration : DATAFILES_ID COLON word"""
    p[0] = Node("data_files_declaration", value=p[3].value)

def p_data_files_stmts(p):
    """data_files_stmts : data_files_stmts data_files_stmt"""
    p[0] = Node("data_files_stmts", children=(p[1].children + [p[2]]))

def p_data_files_stmts_term(p):
    """data_files_stmts : data_files_stmt"""
    p[0] = Node("data_files_stmts", children=[p[1]])

def p_data_files_stmt(p):
    """data_files_stmt : data_files_target
                       | data_files_files
                       | data_files_srcdir
    """
    p[0] = p[1]

def p_data_files_target(p):
    """data_files_target : TARGET_ID COLON anyword"""
    p[0] = Node("target_dir", value=p[3].value)

def p_data_files_srcdir(p):
    """data_files_srcdir : SRCDIR_ID COLON anyword"""
    p[0] = Node("source_dir", value=p[3].value)

def p_data_files_files(p):
    """data_files_files : FILES_ID COLON comma_list"""
    p[0] = Node("files", value=p[3].value)

#---------------------
#   Library section
#---------------------
def p_library(p):
    """library : library_declaration INDENT library_stmts DEDENT
    """
    p[0] = Node("library", children=[p[1]])
    p[0].children.append(p[3])

def p_library_decl_only(p):
    """library : library_declaration
    """
    p[0] = Node("library", children=[p[1]])

def p_library_declaration(p):
    """library_declaration : LIBRARY_ID COLON library_name"""
    p[0] = p[3]

def p_library_name(p):
    """library_name : word
                    |"""
    if len(p) == 1:
        name = "default"
    else:
        name = p[1]
    p[0] = Node("library_name", value=name)

def p_library_stmts(p):
    """library_stmts : library_stmts library_stmt
    """
    children = p[1].children
    children.append(p[2])
    p[0] = Node("library_stmts", children=children)

def p_library_stmts_term(p):
    """library_stmts : library_stmt
    """
    p[0] = Node("library_stmts", children=[p[1]])

def p_library_stmt(p):
    """library_stmt : modules_stmt
                    | packages_stmt
                    | extension_stmt
                    | compiled_library_stmt
                    | build_requires_stmt
                    | install_requires_stmt
                    | conditional_stmt
                    | sub_directory_stmt
    """
    p[0] = p[1]

def p_packages_stmt(p):
    """packages_stmt : PACKAGES_ID COLON comma_list"""
    p[0] = Node("packages", value=p[3].value)

def p_modules_stmt(p):
    """modules_stmt : MODULES_ID COLON comma_list"""
    p[0] = Node("modules", value=p[3].value)

def p_build_requires_stmt(p):
    """build_requires_stmt : BUILD_REQUIRES_ID COLON comma_list"""
    p[0] = Node("build_requires", value=p[3].value)

def p_install_requires_stmt(p):
    """install_requires_stmt : INSTALL_REQUIRES_ID COLON comma_list"""
    p[0] = Node("install_requires", value=p[3].value)

def p_extension_stmt_content(p):
    """extension_stmt : extension_decl INDENT extension_field_stmts DEDENT"""
    p[0] = Node("extension", children=[p[1]])
    p[0].children.append(p[3])

def p_extension_field_stmts(p):
    """extension_field_stmts : extension_field_stmts extension_field_stmt"""
    children = p[1].children
    children.append(p[2])
    p[0] = Node("extension_field_stmts", children=children)

def p_extension_field_stmts_term(p):
    """extension_field_stmts : extension_field_stmt"""
    p[0] = Node("extension_field_stmts", children=[p[1]])

def p_extension_decl(p):
    """extension_decl : EXTENSION_ID COLON anyword"""
    p[0] = Node("extension_declaration", value=p[3].value)

def p_extension_sources(p):
    """extension_field_stmt : SOURCES_ID COLON comma_list"""
    p[0] = Node("sources", value=p[3].value)

def p_extension_include_dirs(p):
    """extension_field_stmt : INCLUDE_DIRS_ID COLON comma_list"""
    p[0] = Node("include_dirs", value=p[3].value)

def p_compiled_library_stmt_content(p):
    """compiled_library_stmt : compiled_library_decl INDENT compiled_library_field_stmts DEDENT"""
    p[0] = Node("compiled_library", children=[p[1]])
    p[0].children.append(p[3])

def p_compiled_library_field_stmts(p):
    """compiled_library_field_stmts : compiled_library_field_stmts compiled_library_field_stmt"""
    children = p[1].children
    children.append(p[2])
    p[0] = Node("compiled_library_field_stmts", children=children)

def p_compiled_library_field_stmts_term(p):
    """compiled_library_field_stmts : compiled_library_field_stmt"""
    p[0] = Node("compiled_library_field_stmts", children=[p[1]])

def p_compiled_library_decl(p):
    """compiled_library_decl : COMPILED_LIBRARY_ID COLON anyword"""
    p[0] = Node("compiled_library_declaration", value=p[3].value)

def p_compiled_library_sources(p):
    """compiled_library_field_stmt : SOURCES_ID COLON comma_list"""
    p[0] = Node("sources", value=p[3].value)

def p_compiled_library_include_dirs(p):
    """compiled_library_field_stmt : INCLUDE_DIRS_ID COLON comma_list"""
    p[0] = Node("include_dirs", value=p[3].value)

def p_sub_directory_stmt(p):
    """sub_directory_stmt : SUB_DIRECTORY_ID COLON word"""
    p[0] = Node("sub_directory", value=p[3].value)

#---------------------
# Conditional handling
#---------------------
def p_in_conditional_stmts(p):
    """in_conditional_stmts : library_stmts
                            | path_stmts
    """
    p[0] = p[1]

def p_conditional_if_only(p):
    """conditional_stmt : IF test COLON INDENT in_conditional_stmts DEDENT"""
    p[0] = Node("conditional", value=p[2], children=[p[5]])

def p_conditional_if_else(p):
    """conditional_stmt : IF test COLON INDENT in_conditional_stmts DEDENT \
                          ELSE COLON INDENT in_conditional_stmts DEDENT
    """
    p[0] = Node("conditional", value=p[2], children=[p[5], p[10]])

def p_test(p):
    """test : bool
            | os_var
            | flag_var"""
    p[0] = p[1]

def p_os_var(p):
    """os_var : OS_OP LPAR word RPAR"""
    p[0] = Node("osvar", value=p[3])

def p_flag_var(p):
    """flag_var : FLAG_OP LPAR word RPAR"""
    p[0] = Node("flagvar", value=p[3])

def p_not_flag_var(p):
    """flag_var : NOT_OP FLAG_OP LPAR word RPAR"""
    p[0] = Node("not_flagvar", value=p[4])

def p_cond_expr_true(p):
    """bool : TRUE"""
    p[0] = Node("bool", value=True)

def p_cond_expr_true_not(p):
    """bool : NOT_OP FALSE"""
    p[0] = Node("bool", value=True)

def p_cond_expr_false(p):
    """bool : FALSE"""
    p[0] = Node("bool", value=False)

def p_cond_expr_false_not(p):
    """bool : NOT_OP TRUE"""
    p[0] = Node("bool", value=False)

#---------------------
#   Path section
#---------------------
def p_path(p):
    """path : path_declaration INDENT path_stmts DEDENT"""
    p[0] = Node("path", children=[p[1], p[3]])

def p_path_declaration(p):
    """path_declaration : PATH_ID COLON anyword"""
    p[0] = Node("path_declaration", value=p[3].value)

def p_path_stmts(p):
    """path_stmts : path_stmts path_stmt"""
    p[0] = Node("path_stmts", children=(p[1].children + [p[2]]))

def p_path_stmts_term(p):
    """path_stmts : path_stmt"""
    p[0] = Node("path_stmts", children=[p[1]])

def p_path_stmt(p):
    """path_stmt : path_description
                 | path_default
                 | conditional_stmt"""
    p[0] = p[1]

def p_path_description(p):
    """path_description : DESCRIPTION_ID COLON single_line_value"""
    #"""path_description : meta_description_stmt"""
    p[0] = Node("path_description", value=p[3])

def p_path_default(p):
    """path_default : DEFAULT_ID COLON anyword"""
    p[0] = Node("path_default", value=p[3].value)

#------------------
#   Flag section
#------------------
def p_flag(p):
    """flag : flag_declaration INDENT flag_stmts DEDENT"""
    p[0] = Node("flag", children=[p[1], p[3]])

def p_flag_declaration(p):
    """flag_declaration : FLAG_ID COLON anyword"""
    p[0] = Node("flag_declaration", value=p[3].value)

def p_flag_stmts(p):
    """flag_stmts : flag_stmts flag_stmt"""
    p[0] = Node("flag_stmts", children=(p[1].children + [p[2]]))

def p_flag_stmts_term(p):
    """flag_stmts : flag_stmt"""
    p[0] = Node("flag_stmts", children=[p[1]])

def p_flag_stmt(p):
    """flag_stmt : flag_description
                 | flag_default"""
    p[0] = p[1]

def p_flag_description(p):
    """flag_description : DESCRIPTION_ID COLON single_line_value"""
    #"""flag_description : meta_description_stmt"""
    p[0] = Node("flag_description", value=p[3])

def p_flag_default(p):
    """flag_default : DEFAULT_ID COLON anyword"""
    p[0] = Node("flag_default", value=p[3].value)

#----------------------
#  Executable section
#----------------------
def p_executable(p):
    """exec : exec_decl INDENT exec_stmts DEDENT"""
    p[0] = Node("executable", children=[p[1], p[3]])

def p_exec_declaration(p):
    """exec_decl : EXECUTABLE_ID COLON anyword"""
    p[0] = Node("exec_name", value=p[3].value)

def p_exec_stmts(p):
    """exec_stmts : exec_stmts exec_stmt"""
    p[0] = Node("exec_stmts", children=(p[1].children + [p[2]]))

def p_exec_stmts_term(p):
    """exec_stmts : exec_stmt"""
    p[0] = Node("exec_stmts", children=[p[1]])

def p_exec_stmt(p):
    """exec_stmt : function
                 | module"""
    p[0] = p[1]

def p_exec_module(p):
    """module : MODULE_ID COLON anyword"""
    p[0] = Node("module", value=p[3].value)

def p_exec_function(p):
    """function : FUNCTION_ID COLON anyword"""
    p[0] = Node("function", value=p[3].value)

#-----------------------
#   Literal handling
#-----------------------

def p_comma_list_indented(p):
    """comma_list : indented_comma_list 
    """
    p[0] = Node("comma_list", value=p[1].value)

def p_comma_list(p):
    """comma_list : comma_words
    """
    p[0] = Node("comma_list", value=p[1].value)

def p_indented_comma_list(p):
    """indented_comma_list : comma_words COMMA INDENT comma_words DEDENT
    """
    p[0] = p[1]
    p[0].value.extend(p[4].value)

def p_indented_comma_list_term(p):
    """indented_comma_list : INDENT comma_words DEDENT
    """
    p[0] = p[2]

def p_comma_words(p):
    """comma_words : comma_words COMMA anyword_comma_list
    """
    p[0] = p[1]
    p[0].value.append(p[3].value)

def p_comma_words_term(p):
    """comma_words : anyword_comma_list
    """
    p[0] = Node("comma_words", value=[p[1].value])

def p_single_line(p):
    """single_line_value : WS single_line"""
    p[0] = p[2]

def p_single_line_no_space(p):
    """single_line_value : single_line"""
    p[0] = p[1]

def p_single_line_string(p):
    """single_line : single_line literal"""
    p[0] = p[1] + [p[2]]

def p_single_line_string_term(p):
    """single_line : literal_no_space"""
    p[0] = [p[1]]

def p_meta_description_stmt_start_same_line(p):
    """meta_description_stmt : description_decl single_line_newline \
                               INDENT multi_stmts DEDENT
    """
    p[0] = Node("description", value=(p[2].value + p[4]))

def p_meta_description_stmt_indented_block(p):
    """meta_description_stmt : description_decl NEWLINE \
                               INDENT multi_stmts DEDENT
    """
    p[0] = Node("description", value=p[4])

def p_meta_description_stmt_indented_block2(p):
    """meta_description_stmt : description_decl WS NEWLINE \
                               INDENT multi_stmts DEDENT
    """
    p[0] = Node("description", value=p[5])

def p_meta_description_stmt_single(p):
    """meta_description_stmt : description_decl single_line_value
    """
    p[0] = Node("description", value=p[2])

def p_single_line_newline(p):
    """single_line_newline : single_line_value newline
    """
    p[0] = Node("single_line_newline", value=(p[1] + [p[2]]))

def p_description_decl(p):
    """description_decl : DESCRIPTION_ID COLON"""
    p[0] = Node("description_decl")

def p_indented_block(p):
    """indented_block : indent indented_block_value"""
    p[0] = [p[1]] + p[2]

def p_indented_block_value(p):
    """indented_block_value : multi_stmts dedent"""
    p[0] = p[1] + [p[2]]

def p_multi_stmts(p):
    """multi_stmts : multi_stmt multi_stmts"""
    p[0] = p[1] + p[2]

def p_multi_stmts_term(p):
    """multi_stmts : multi_stmt"""
    p[0] = p[1]

def p_multi_stmt_ind_block(p):
    """multi_stmt : indented_block"""
    p[0] = p[1]

def p_multi_stmt_term(p):
    """multi_stmt : multi_literal"""
    p[0] = [p[1]]

def p_newline(p):
    """newline : NEWLINE"""
    p[0] = Node("newline", value=p[1])

# anyword groks any character stream without space|newline
def p_anyword(p):
    """anyword : anyword literal"""
    p[0] = Node("anyword", value=(p[1].value + p[2].value))

def p_anyword_term(p):
    """anyword : literal"""
    p[0] = p[1]

def p_anyword_comma_list(p):
    """anyword_comma_list : anyword_comma_list anytoken_no_comma"""
    p[0] = Node("anyword_comma_list", value=(p[1].value + p[2].value))

def p_anyword_comma_list_term(p):
    """anyword_comma_list : anytoken_no_comma"""
    p[0] = Node("anyword_comma_list", value=p[1].value)

# Any token but whitespace, newline and comma
def p_anytoken_no_comma(p):
    """anytoken_no_comma : WORD
                         | COLON
                         | DOUBLE_COLON
                         | LPAR
                         | RPAR
                         | LESS
                         | SLASH
                         | SHARP
                         | EQUAL
                         | GREATER
    """
    p[0] = Node("anytoken", value=p[1])

def p_literal_no_space(p):
    """literal_no_space : anytoken_no_comma
    """
    p[0] = Node("literal", value=p[1].value)

def p_literal_no_space_term(p):
    """literal_no_space : COMMA
    """
    p[0] = Node("literal", value=p[1])

def p_literal(p):
    """literal : literal_no_space
    """
    p[0] = Node("literal", value=p[1].value)

def p_literal_term(p):
    """literal : WS
    """
    p[0] = Node("literal", value=p[1])

def p_multi_literal(p):
    """multi_literal : literal
    """
    p[0] = Node("multi_literal", value=p[1].value)

def p_multi_literal2(p):
    """multi_literal : newline
    """
    p[0] = p[1]

def p_indent(p):
    """indent : INDENT
    """
    p[0] = Node("indent", value=p[1])

def p_dedent(p):
    """dedent : DEDENT
    """
    p[0] = Node("dedent", value=p[1])

def p_word(p):
    """word : WORD"""
    p[0] = Node("word", value=p[1])

def p_version(p):
    """version : WORD"""
    p[0] = Node("version", value=p[1])

def p_error(p):
    if p is None:
        raise InternalBentoError("Unknown parsing error (parser/lexer bug ? Please report this with your bento.info)")
    else:
        msg = "yacc: Syntax error at line %d, Token(%s, %r)" % \
                (p.lineno, p.type, p.value)
        if hasattr(p.lexer, "lexdata"):
            data = p.lexer.lexdata.splitlines()
            msg += "\n\t%r" % (data[p.lineno-1],)
        raise ParseError(msg, p)

#def error_msg(p, error_msg):
#    if p is not None:
#        msg = ["Syntax error at line number %d, token %s (%r)" % \
#               (p.lineno, p.type, p.value)]
#        if error_msg is not None:
#            msg += ["    %s" % error_msg]
#        if hasattr(p.lexer, "lexdata"):
#            data = p.lexer.lexdata.splitlines()
#            msg += ["    Line %d -> %r" % (p.lineno, data[p.lineno-1])]
#        else:
#            msg += ["    Line %d" % (p.lineno)]
#        return "\n".join(msg)
#    return "Unhandled token"



import os

from pprint import \
        pprint

from bento.errors \
    import \
        UsageException, ParseError
from bento.commands.core import \
        Command, Option
from bento.parser.misc \
    import \
        build_ast_from_data
from bento.utils.utils \
    import \
        extract_exception

class ParseCommand(Command):
    long_descr = """\
Purpose: query the given package description file (debugging tool)
Usage:   bentomaker parse [OPTIONS]"""
    short_descr = "parse the package description file."
    common_options = Command.common_options + [
        Option("-f", "--flags", action="store_true",
               help="print flags variables"),
        Option("-p", "--path", action="store_true",
               help="print paths variables"),
        Option("-m", "--meta-field", dest="meta_field",
               help="print given meta field")]

    def run(self, ctx):
        argv = ctx.command_argv
        p = ctx.options_context.parser
        o, a = p.parse_args(argv)
        if o.help:
            p.print_help()
            return

        if len(a) < 1:
            filename = "bento.info"
        else:
            filename = a[0]

        if not os.path.exists(filename):
            raise UsageException("%s: error: file %s does not exist" % "bentomaker")

        f = open(filename, "r")
        try:
            data = f.read()
            try:
                parsed = build_ast_from_data(data)
            except ParseError:
                e = extract_exception()
                msg = "Error while parsing file %s\n" % filename
                e.args = (msg,) +  e.args
                raise e
            if o.flags:
                try:
                    flags = parsed["flags"]
                    for flag in flags:
                        print(flags[flag])
                except KeyError:
                    pass
            elif o.path:
                try:
                    paths = parsed["paths"]
                    for path in paths:
                        print(paths[path])
                except KeyError:
                    pass
            elif o.meta_field:
                try:
                    print(parsed[o.meta_field])
                except KeyError:
                    raise ValueError("Field %s not found in metadata" % o.meta_field)
            else:
                pprint(parsed)
        finally:
            f.close()




import sys

import six

class Peeker(object):
    """Generator to enable "peeking" the next item

    Parameters
    ----------
    it : iterator
        iterator which we one to peek in
    dummy :
        if not None, will be returned by peek if the iterator is empty

    Example
    -------
    >>> a = [1, 2, 3, 4]
    >>> peeker = Peeker(a)
    >>> for i in peeker:
    >>>     try:
    >>>         next = peeker.peek()
    >>>         print "Next to %d is %d" % (i, next)
    >>>     except StopIteration:
    >>>         print "End of stream", i
    """
    def __init__(self, it, dummy=None):
        self._it = iter(it)
        self._cache = None
        if dummy is None:
            self.peek = self._peek_no_dummy
        else:
            self.peek = self._peek_dummy
        self._dummy = dummy

    def __next__(self):
        return self.next()

    def next(self):
        if self._cache:
            i = self._cache
            self._cache = None
            return i
        else:
            return six.advance_iterator(self._it)
        #self._cache = None
        #return i

    def _peek_dummy(self):
        if self._cache:
            return self._cache
        else:
            try:
                i = six.advance_iterator(self._it)
            except StopIteration:
                return self._dummy
            self._cache = i
            return i

    def _peek_no_dummy(self):
        if self._cache:
            return self._cache
        else:
            i = six.advance_iterator(self._it)
            self._cache = i
            return i

    def __iter__(self):
        return self

class BackwardGenerator(object):
    def __init__(self, gen):
        self._gen = gen
        self._cache = []
        self._previous = None

    def next(self):
        c = six.advance_iterator(self._gen)
        if len(self._cache) == 2:
            old, new = self._cache
            self._cache = [new]
        self._cache.append(c)
        return c

    def __next__(self):
        return self.next()

    def previous(self):
        if len(self._cache) < 2:
            raise ValueError()
        return self._cache[0]

    def __iter__(self):
        return self

def print_tokens_simple(lexer):
    while True:
        tok = lexer.token()
        if not tok:
            break
        print(tok)



import os
import tempfile
import sys

from bento.compat.api.moves \
    import \
        unittest
from bento.core.pkg_objects \
    import \
        DataFiles, Executable

class TestDataFiles(unittest.TestCase):
    def test_simple(self):
        data = DataFiles("data", files=["foo.c"])
        self.assertEqual(data.name, "data")
        self.assertEqual(data.files, ["foo.c"])
        self.assertEqual(data.source_dir, ".")
        self.assertEqual(data.target_dir, "$sitedir")

    def test_from_dict(self):
        parsed_dict = {"name": "data",
                       "files": ["foo.c", "yo.c"], "target_dir": "foo"}
        data = DataFiles.from_parse_dict(parsed_dict)
        self.assertEqual(data.name, "data")
        self.assertEqual(data.files, ["foo.c", "yo.c"])
        self.assertEqual(data.target_dir, "foo")
    # TODO: test with a populated temp dir

class TestExecutable(unittest.TestCase):
    def test_basic(self):
        exe = Executable.from_representation("foo = core:main")
        self.assertEqual(exe.name, "foo")
        self.assertEqual(exe.module, "core")
        self.assertEqual(exe.function, "main")

        self.assertEqual(exe.full_representation(), "foo = core:main")
